@@ -21,6 +21,7 @@
 #include <dbus/dbus.h>
 #include <dbus/dbus-glib.h>
 #include <libhal.h>
+#include <signal.h>
 
 #include "gvm.h"
 
@@ -37,7 +38,7 @@
 # define N_(String) (String)
 #endif
 
-//#define GVM_DEBUG
+#define GVM_DEBUG
 #ifdef GVM_DEBUG
 # define dbg(fmt,arg...) fprintf(stderr, "%s/%d: " fmt,__FILE__,__LINE__,##arg)
 #else
@@ -46,12 +47,17 @@
 
 #define warn(fmt,arg...) g_warning("%s/%d: " fmt,__FILE__,__LINE__,##arg)
 
-#define BIN_MOUNT	"/bin/mount"	/* what we mount with */
-#define NAUTILUS_COMMAND "/usr/bin/nautilus -n --no-desktop %m"
+#define NAUTILUS_COMMAND	 BIN_NAUTILUS" -n --no-desktop %m"
 
 static struct gvm_configuration config;
 static LibHalContext *hal_ctx;
 
+/** List of UDI's for volumes mounted by g-v-m that we need to apply policy to*/
+static GSList *mounted_volumes_policy_queue = NULL;
+
+/** List of UDI's of all volumes mounted during the lifetime of the program */
+static GSList *all_mounted_volumes = NULL;
+
 /*
  * gvm_load_config - synchronize gconf => config structure
  */
@@ -136,11 +142,13 @@ gvm_init_config (void)
 }
 
 /*
- * gvm_run_command - run the given command, replacing %d with the device node
- * and %m with the given path
+ * gvm_run_command - run the given command, replacing %d with the device node,
+ * %h with the HAL UDI and %m with the given path. Bails out if the command
+ * requires a given %d,%h or %m and this is NULL.
  */
 static void
-gvm_run_command (const char *device, const char *command, const char *path)
+gvm_run_command (const char *device, const char *command, 
+		 const char *path, const char *udi)
 {
 	char *argv[4];
 	gchar *new_command;
@@ -148,23 +156,37 @@ gvm_run_command (const char *device, const char *command, const char *path)
 	GString *exec = g_string_new (NULL);
 	char *p, *q;
 
-	/* perform s/%d/device/ and s/%m/path/ */
+	/* perform s/%d/device/, s/%m/path/ and s/%h/udi/ */
 	new_command = g_strdup (command);
 	q = new_command;
 	p = new_command;
 	while ((p = strchr (p, '%')) != NULL) {
 		if (*(p + 1) == 'd') {
+			if (device == NULL)
+				goto error;
 			*p = '\0';
 			g_string_append (exec, q);
 			g_string_append (exec, device);
 			q = p + 2;
 			p = p + 2;
 		} else if (*(p + 1) == 'm') {
+			if (path == NULL)
+				goto error;
 			*p = '\0';
 			g_string_append (exec, q);
 			g_string_append (exec, path);
 			q = p + 2;
 			p = p + 2;
+		} else if (*(p + 1) == 'h') {
+			if (udi == NULL)
+				goto error;
+			*p = '\0';
+			g_string_append (exec, q);
+			g_string_append (exec, "\"");
+			g_string_append (exec, udi);
+			g_string_append (exec, "\"");
+			q = p + 2;
+			p = p + 2;
 		}
 	}
 	g_string_append (exec, q);
@@ -181,6 +203,13 @@ gvm_run_command (const char *device, const char *command, const char *path)
 
 	g_string_free (exec, TRUE);
 	g_free (new_command);
+	return;
+
+error:
+	warn ("command '%s' required unavailable parameter; " 
+	      "%%d='%s' %%m='%s' %%h='%s'\n", command, device, path, udi);
+	g_string_free (exec, TRUE);
+	g_free (new_command);	
 }
 
 /*
@@ -220,7 +249,7 @@ gvm_ask_autorun (const char *path)
  * Returns TRUE if this was a Video DVD and FALSE otherwise.
  */
 static gboolean
-gvm_check_dvd (const char *device, const char *mount_point)
+gvm_check_dvd (const char *device, const char *mount_point, const char *udi)
 {
 	char *path;
 	gboolean retval;
@@ -239,7 +268,7 @@ gvm_check_dvd (const char *device, const char *mount_point)
 
 	if (retval && config.autoplay_dvd)
 		gvm_run_command (device, config.autoplay_dvd_command,
-				 mount_point);
+				 mount_point, udi);
 
 	return retval;
 }
@@ -291,7 +320,7 @@ gvm_check_photos (const char *udi, const char *device, const char *mount_point)
 
 		if (action == IMPORT)
 			gvm_run_command (device, config.autophoto_command,
-					 dcim_path);
+					 dcim_path, udi);
 	}
 
 out:
@@ -299,6 +328,47 @@ out:
 	return retval;
 }
 
+
+/*
+ * gvm_check_camera - check if this device is a digital camera. If it is, then
+ * ask the user if he wants to import the photos.
+ *
+ * Returns TRUE if this was a camera, FALSE otherwise
+ */
+static gboolean
+gvm_check_camera (const char *udi)
+{
+	enum { IMPORT } action = -1;
+	GtkWidget *askme;
+	int retval = FALSE;
+
+	/* see if it's a camera */
+	if (!hal_device_query_capability(hal_ctx, udi, "camera"))
+		goto out;
+
+	retval = TRUE;
+	dbg ("Camera detected: %s\n", udi);
+
+	if (config.autophoto) {
+		askme = gtk_message_dialog_new (NULL, 0, GTK_MESSAGE_QUESTION,
+						GTK_BUTTONS_NONE,
+						_(ASK_PHOTOS_MSG));
+		gtk_dialog_add_buttons (GTK_DIALOG (askme),
+					GTK_STOCK_CANCEL,
+					GTK_RESPONSE_CANCEL, _("_Import"),
+					IMPORT, NULL);
+		action = gtk_dialog_run (GTK_DIALOG (askme));
+		gtk_widget_destroy (askme);
+
+		if (action == IMPORT)
+			gvm_run_command (NULL, config.autophoto_command,
+					 NULL, udi);
+	}
+
+out:
+	return retval;
+}
+
 /*
  * gvm_device_autorun - automatically execute stuff on the given UDI
  *
@@ -323,7 +393,7 @@ gvm_device_autorun (const char *udi)
 		goto out;
 	}
 
-	if (gvm_check_dvd (device, mount_point))
+	if (gvm_check_dvd (device, mount_point, udi))
 		goto out;
 
 	if (gvm_check_photos (udi, device, mount_point))
@@ -368,7 +438,7 @@ gvm_device_autorun (const char *udi)
 	}
 	
 	if ((config.autobrowse == TRUE) && (autorun_succeeded == FALSE)) {
-		gvm_run_command (device, NAUTILUS_COMMAND, mount_point);
+		gvm_run_command (device, NAUTILUS_COMMAND, mount_point, udi);
 	}
 
 out:
@@ -381,21 +451,55 @@ out:
  *
  * Note that this requires that the given device node is in /etc/fstab.  This
  * is intentional.
+ *
+ * @return TRUE iff the mount was succesful
  */
-static void
+static gboolean
 gvm_device_mount (char *device)
 {
 	char *argv[3];
 	GError *error = NULL;
+	gint exit_status;
 
 	argv[0] = BIN_MOUNT;
 	argv[1] = device;
 	argv[2] = NULL;
 
-	g_spawn_async (g_get_home_dir (), argv, NULL, 0, NULL,
-		       NULL, NULL, &error);
-	if (error)
+	if (!g_spawn_sync (g_get_home_dir (), argv, NULL, 0, NULL,
+			   NULL, NULL, NULL, &exit_status, &error)) {
 		warn ("failed to exec " BIN_MOUNT ": %s\n", error->message);
+		return FALSE;
+	}
+
+	return (exit_status == 0);
+}
+
+/*
+ * gvm_device_unmount - use BIN_UMOUNT to unmount the given device node.
+ *
+ * Note that this requires that the given device node is in /etc/fstab.  This
+ * is intentional.
+ *
+ * @return TRUE iff the mount was succesful
+ */
+static gboolean
+gvm_device_unmount (char *device)
+{
+	char *argv[3];
+	GError *error = NULL;
+	gint exit_status;
+
+	argv[0] = BIN_UMOUNT;
+	argv[1] = device;
+	argv[2] = NULL;
+
+	if (!g_spawn_sync (g_get_home_dir (), argv, NULL, 0, NULL,
+			   NULL, NULL, NULL, &exit_status, &error)) {
+		warn ("failed to exec " BIN_MOUNT ": %s\n", error->message);
+		return FALSE;
+	}
+
+	return (exit_status == 0);
 }
 
 /*
@@ -403,11 +507,11 @@ gvm_device_mount (char *device)
  * the given device node
  */
 static void
-gvm_run_cdplayer (const char *device, const char *mount_point)
+gvm_run_cdplayer (const char *device, const char *mount_point, const char *udi)
 {
 	if (config.autoplay_cda)
 		gvm_run_command (device, config.autoplay_cda_command,
-				 mount_point);
+				 mount_point, udi);
 }
 
 #define ASK_MIXED_MSG	"This CD has both audio tracks and data files.\n" \
@@ -453,9 +557,10 @@ gvm_ask_mixed (const char *udi)
 	switch (action) {
 	case MOUNT:
 		gvm_device_mount (device);
+		mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
 		break;
 	case PLAY:
-		gvm_run_cdplayer (device, device);
+		gvm_run_cdplayer (device, device, udi);
 		break;
 	default:
 		break;
@@ -471,10 +576,10 @@ out:
  * given device node, if so configured
  */
 static void
-gvm_run_cdburner (const char *device, const char *mount)
+gvm_run_cdburner (const char *device, const char *mount, const char *udi)
 {
 	if (config.autoburn_cdr)
-		gvm_run_command (device, config.autoburn_cdr_command, mount);
+		gvm_run_command (device, config.autoburn_cdr_command, mount, udi);
 }
 
 /*
@@ -523,15 +628,17 @@ gvm_cdrom_policy (const char *udi)
 	}
 
 	if (has_audio && (!has_data)) {
-		gvm_run_cdplayer (device, device);
+		gvm_run_cdplayer (device, device, udi);
 	} else if (has_audio && has_data) {
 		gvm_ask_mixed (udi);
 	} else if (has_data) {
-		if (config.automount_media)
+		if (config.automount_media) {
 			gvm_device_mount (device);
+			mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
+		}
 	} else if (is_blank) {
 		if (gvm_device_is_writer (drive_udi))
-			gvm_run_cdburner (device, device);
+			gvm_run_cdburner (device, device, udi);
 	}
 
 	/** @todo enforce policy for all the new disc types now supported */
@@ -552,10 +659,21 @@ out:
  * At the moment, we only handle CD-ROM and DVD drives.
  */
 static void
-gvm_media_changed (const char *udi, const char *storage_device)
+gvm_media_changed (const char *udi, const char *storage_device, 
+		   const char *device)
 {
 	char *media_type;
 
+	/* Refuse to enforce policy on removable media if drive is locked */
+	if (hal_device_property_exists (
+		    hal_ctx, storage_device, "info.locked") &&
+	    hal_device_get_property_bool (
+		    hal_ctx, storage_device, "info.locked")) {
+		dbg ("Drive with udi %s is locked through hal; "
+		     "skipping policy\n", storage_device);
+		return;
+	}
+
 	/*
 	 * Get HAL's interpretation of our media type.  Note that we must check
 	 * the storage device and not this UDI
@@ -567,10 +685,16 @@ gvm_media_changed (const char *udi, const char *storage_device)
 		return;
 	}
 
-	if (!g_strcasecmp (media_type, "cdrom"))
+	if (!g_strcasecmp (media_type, "cdrom")) {
 		gvm_cdrom_policy (udi);
-
-	/* other media_types go here */
+	} else {
+		dbg ("Added: %s\n", device);
+		
+		if (config.automount_drives) {
+			gvm_device_mount (device);
+			mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
+		}
+	}
 
 	hal_free_string (media_type);
 }
@@ -588,6 +712,8 @@ hal_device_added (LibHalContext *ctx __attribute__((__unused__)),
 
 	dbg ("New Device: %s\n", udi);
 
+	gvm_check_camera (udi);
+
 	if (!hal_device_query_capability(hal_ctx, udi, "block"))
 		goto out;
 	
@@ -621,15 +747,17 @@ hal_device_added (LibHalContext *ctx __attribute__((__unused__)),
 					  "storage.removable")) {
 		/* we handle media change events separately */
 		dbg ("Changed: %s\n", device);
-		gvm_media_changed (udi, storage_device);
+		gvm_media_changed (udi, storage_device, device);
 		goto out;
 	}
 	
 	/* folks, we have a new device! */
 	dbg ("Added: %s\n", device);
 	
-	if (config.automount_drives)
+	if (config.automount_drives) {
 		gvm_device_mount (device);
+		mounted_volumes_policy_queue = g_slist_append (mounted_volumes_policy_queue, g_strdup (udi));
+	}
 	
 out:
 	hal_free_string (device);
@@ -688,16 +816,50 @@ hal_property_modified (LibHalContext *ctx __attribute__((__unused__)),
 		       dbus_bool_t is_removed __attribute__((__unused__)), 
 		       dbus_bool_t is_added __attribute__((__unused__)))
 {
-	if (!g_strcasecmp (key, "volume.is_mounted")) {
-		dbus_bool_t val;
-		
-		val = hal_device_get_property_bool (hal_ctx, udi, key);
-		if (val == TRUE) {
-			dbg ("Mounted: %s\n", udi);
+	dbus_bool_t val;
+	GSList *i;
+	GSList *next;
+
+	if (g_strcasecmp (key, "volume.is_mounted") != 0)
+		return;
+	
+	val = hal_device_get_property_bool (hal_ctx, udi, key);
+	if (val == TRUE) {
+		GSList *policy_udi;
+
+		dbg ("Mounted: %s\n", udi);
+
+		/* add to list of all volumes mounted during lifetime */
+		all_mounted_volumes = g_slist_append (all_mounted_volumes,
+						      g_strdup (udi));
+
+		policy_udi = g_slist_find_custom (mounted_volumes_policy_queue, 
+						  udi, 
+						  g_ascii_strcasecmp);
+		if (policy_udi != NULL) {
+			g_free (policy_udi->data);
+			mounted_volumes_policy_queue = g_slist_delete_link (mounted_volumes_policy_queue, 
+									    policy_udi);
 			gvm_device_autorun (udi);
-		} else
-			dbg ("Unmounted: %s\n", udi);
-	}	
+		}
+
+	} else {
+		dbg ("Unmounted: %s\n", udi);
+
+		/* remove from list of all volumes mounted during lifetime */
+
+		for (i=all_mounted_volumes; i != NULL; i = next) {
+			next = g_slist_next (i);
+			if (strcmp (udi, (const char *)i->data) == 0) {
+				g_free (i->data);
+				all_mounted_volumes = 
+					g_slist_delete_link (
+						all_mounted_volumes, i);
+				break;
+			}
+		}
+
+	}
 }
 
 /** Invoked when a device in the GDL emits a condition that cannot be
@@ -714,29 +876,6 @@ hal_device_condition (LibHalContext *ctx __attribute__((__unused__)),
 		      const char *condition_name __attribute__((__unused__)),
 		      DBusMessage * message __attribute__((__unused__)))
 {
-	if (!g_strcasecmp (condition_name, "EjectPressed")) {		
-		char *argv[3];
-		GError *error = NULL;
-		char *device;
-
-		device = hal_device_get_property_string (ctx, udi,
-							 "block.device");
-		if (!device)
-			warn ("cannot get block.device\n");
-		else if (config.eject_command) {
-			argv[0] = config.eject_command;
-			argv[1] = device;
-			argv[2] = NULL;
-
-			g_spawn_async (g_get_home_dir (), argv, NULL, 0, NULL,
-				       NULL, NULL, &error);
-			if (error)
-				warn ("failed to exec %s: %s\n",
-				      config.eject_command, error->message);
-		}
-
-		hal_free_string (device);
-	}
 }
 
 /** Invoked by libhal for integration with our mainloop. 
@@ -793,6 +932,141 @@ gvm_do_hal_init (LibHalFunctions *functions)
 	return ctx;
 }
 
+/** Attempt to mount all volumes; should be called on startup.
+ *
+ *  @param  ctx                 LibHal context
+ */
+static void mount_all (LibHalContext *ctx)
+{
+	int i;
+	int num_volumes;
+	char **volumes;
+	char *udi;
+	char *device_file;
+
+	if (!config.automount_media)
+		return;
+
+	volumes = hal_find_device_by_capability (ctx, "volume", &num_volumes);
+	for (i = 0; i < num_volumes; i++) {
+		udi = volumes [i];
+
+		/* don't attempt to mount already mounted volumes */
+		if (!hal_device_property_exists (ctx, udi, 
+						"volume.is_mounted") ||
+		    hal_device_get_property_bool (ctx, udi, 
+						  "volume.is_mounted"))
+			continue;
+
+		/* only mount if the block device got a sensible filesystem */
+		if (!hal_device_property_exists (ctx, udi, 
+						"volume.fsusage") ||
+		    strcmp (hal_device_get_property_string (ctx, udi, 
+							    "volume.fsusage"), 
+			    "filesystem") != 0)
+			continue;
+
+		device_file = hal_device_get_property_string (ctx, udi,
+							      "block.device");
+
+		if (device_file != NULL ) {
+
+			dbg ("mount_all: mounting %s\n", device_file);
+
+			gvm_device_mount (device_file);
+
+			hal_free_string (device_file);
+		} else
+			warn ("no device_file for udi=%s\n", udi);
+	}
+
+	hal_free_string_array (volumes);
+}
+
+/** Unmount all volumes that were mounted during the lifetime of this
+ *  g-v-m instance
+ *
+ *  @param  ctx                 LibHal context
+ */
+static void
+unmount_all (LibHalContext *ctx)
+{
+	GSList *i;
+	char *device_file;
+	char *udi;
+
+	dbg ("unmounting all volumes that we saw mounted in our life\n");
+
+	for (i = all_mounted_volumes; i != NULL; i = g_slist_next (i)) {
+
+		udi = i->data;
+
+		device_file = hal_device_get_property_string (ctx, udi,
+							      "block.device");
+		if (device_file != NULL ) {
+
+			dbg ("unmount_all: unmounting %s\n", device_file);
+
+			gvm_device_unmount (device_file);
+			hal_free_string (device_file);
+		} else {
+			warn ("no device_file for udi=%s\n", udi);
+		}
+	}
+}
+
+
+static int sigterm_unix_signal_pipe_fds[2];
+static GIOChannel *sigterm_iochn;
+
+static void 
+handle_sigterm (int value)
+{
+	static char marker[1] = {'S'};
+
+	/* write a 'S' character to the other end to tell about
+	 * the signal. Note that 'the other end' is a GIOChannel thingy
+	 * that is only called from the mainloop - thus this is how we
+	 * defer this since UNIX signal handlers are evil
+	 *
+	 * Oh, and write(2) is indeed reentrant */
+	write (sigterm_unix_signal_pipe_fds[1], marker, 1);
+}
+
+static gboolean
+sigterm_iochn_data (GIOChannel *source, 
+		    GIOCondition condition, 
+		    gpointer user_data)
+{
+	GError *err = NULL;
+	gchar data[1];
+	gsize bytes_read;
+
+	/* Empty the pipe */
+	if (G_IO_STATUS_NORMAL != 
+	    g_io_channel_read_chars (source, data, 1, &bytes_read, &err)) {
+		warn ("Error emptying callout notify pipe: %s",
+		      err->message);
+		g_error_free (err);
+		goto out;
+	}
+
+	dbg ("Received SIGTERM, initiating shutdown\n");
+	unmount_all (hal_ctx);
+	gtk_main_quit();
+
+out:
+	return TRUE;
+}
+
+static void
+gvm_die (GnomeClient *client, gpointer user_data) 
+{
+	dbg ("Received 'die', initiating shutdown\n");
+	unmount_all (hal_ctx);
+	gtk_main_quit ();
+}
+
 int
 main (int argc, char *argv[])
 {
@@ -822,7 +1096,7 @@ main (int argc, char *argv[])
 	}
 
 	gtk_signal_connect (GTK_OBJECT (client), "die",
-			    GTK_SIGNAL_FUNC (gtk_main_quit), NULL);
+			    GTK_SIGNAL_FUNC (gvm_die), NULL);
 
 	hal_ctx = gvm_do_hal_init (&hal_functions);
 	if (!hal_ctx)
@@ -830,6 +1104,21 @@ main (int argc, char *argv[])
 
 	gvm_init_config ();
 
+	/* SIGTERM handling via pipes  */
+	if (pipe (sigterm_unix_signal_pipe_fds) != 0) {
+		warn ("Could not setup pipe, errno=%d", errno);
+		return 1;
+	}
+	sigterm_iochn = g_io_channel_unix_new (sigterm_unix_signal_pipe_fds[0]);
+	if (sigterm_iochn == NULL) {
+		warn ("Could not create GIOChannel");
+		return 1;
+	}
+	g_io_add_watch (sigterm_iochn, G_IO_IN, sigterm_iochn_data, NULL);
+	signal (SIGTERM, handle_sigterm);
+
+	mount_all (hal_ctx);
+
 	gtk_main ();
 
 	return 0;
